#extension GL_ARB_gpu_shader5 : enable
#extension GL_ARB_shader_image_load_store : enable
#extension GL_ARB_shading_language_packing : enable

flat varying vec3 ambientUp;
flat varying vec3 ambientDown;
flat varying vec3 ambientLeft;
flat varying vec3 ambientRight;
flat varying vec3 ambientB;
flat varying vec3 ambientF;
flat varying float ambientA;
varying vec2 texcoord;

#include "/lib/light.glsl"

#ifdef LPV_ENABLED
    #include "/lib/lpv/lpv_common.glsl"
    #include "/lib/lpv/lpv_render.glsl"
#endif

#if SHADOW_FILTERING == 2
    #include "/lib/shadows/shadowfunc_advanced.glsl"

#else
    #include "/lib/shadows/shadowfunc_simple.glsl"
#endif
#include "/lib/shadows/shadowfunc_no_shadow.glsl"

#if SSAO == 2
    #include "/lib/rtao/rtao_lib.glsl"
#elif SSAO == 1
    #include "/lib/rtao/rtao_lib_simple.glsl"
#endif

// Function to calculate the torch lighting with optional LPV support
vec3 CalculateTorch(
    vec2 lightmap,
    vec3 normal,
    vec3 fragpos,
    float ao)
{
    // Cache lightmap.x to avoid redundant swizzles.
    float lm = lightmap.x;

    // Initialize torch color.
    vec3 torch = vec3(3.000, 1.382, 0.408);

    // Apply lightmap power.
    float lightmapPow = lm * lm * lm * lm;
    float lightFactor = lightmapPow + lm * 0.025;
    torch *= lightFactor;
    torch = clamp(torch, 0.0, 4.0) * 2.0;

#ifdef LPV_ENABLED
    // Precompute normal offset multiplier.
    const float normOffsetScale = LPV_NORMAL_STRENGTH * 0.01;
    vec3 normalOffset = normal * normOffsetScale;

    // Transform fragpos into LPV space.
    vec3 transformedPos = mat3(gbufferModelViewInverse) * fragpos + gbufferModelViewInverse[3].xyz;
    vec3 lpvPos = GetLpvPosition(transformedPos) + normalOffset;

    // Simplify 1.0 + (1.0 - ao) to (2.0 - ao)
    vec3 LpvTorchLight = GetLpvBlockLight(SampleLpvLinear(lpvPos)) * (2.0 - ao);

    // Compute LPV light factor from lightmap and apply.
    float lpvLightFactor = clamp(lm * (lm + 0.025), 0.0, 10.0) * 1.5;
    LpvTorchLight *= lpvLightFactor;

    // Precompute reciprocal of LPV_BLEND_DISTANCE (if LPV_BLEND_DISTANCE is a constant).
    float invBlend = 1.0 / LPV_BLEND_DISTANCE;

    // Compute cubic radius based on LPV position.
    vec3 radiusA = (((LpvSize3 - 1.0) - lpvPos) * invBlend);
    vec3 radiusB = (lpvPos * invBlend);
    vec3 cubicRadius = clamp(min(radiusA, radiusB), 0.0, 1.0);

    // Combine the three components.
    float LpvFadeF = cubicRadius.x * cubicRadius.y * cubicRadius.z;

    // Smooth the fade factor.
    LpvFadeF = 1.0 - pow(1.0 - pow(LpvFadeF, 1.5), 3.0);

    #ifdef DYNAMIC_LIGHTS_COMP
    // Optionally adjust fade factor based on the distance between LPV and torch colors.
    LpvFadeF = clamp(1.0 - pow(distance(LpvTorchLight, torch), 0.25), 0.0, 1.0);
    #endif

    // Blend the base torch with the LPV torch light.
    torch = mix(torch, LpvTorchLight, LpvFadeF);
#endif

    return torch;
}

#include "/lib/materials.glsl"
#include "/lib/reflection_lib_fake.glsl"

float gbf(in vec3 eyePlayerPos, in float far, in float z0)
{
    float distXZ = length((vec3(eyePlayerPos.x, 0.0, eyePlayerPos.z)).xyz);
    float distY = length((vec3(0.0, eyePlayerPos.y, 0.0)).xyz);
    float eyeDist = max(distXZ, distY);
#ifdef DISTANT_HORIZONS
    float borderFogFactor = smoothstep(0.0, dhRenderDistance, eyeDist);
    borderFogFactor *= borderFogFactor * borderFogFactor;
    return borderFogFactor;
#else

    float borderFogFactor = smoothstep(far * 0.5, far, eyeDist);
    borderFogFactor *= borderFogFactor * borderFogFactor;

    return borderFogFactor;

#endif
}

vec3 CalculateWaterCaustics(vec3 screenSpacePosition)
{
    vec3 worldPos = mat3(gbufferModelViewInverse) * screenSpacePosition + cameraPosition;

    float scale = worldPos.y / lightPos.y;
    vec2 pos = worldPos.xz - lightPos.xz * scale;

    vec2 coord0 = fract(pos / 128.0);
    vec2 coord1 = fract(pos / 16.0);

    float caustics = texture(colortex7, coord0).g + texture(colortex7, coord1).g;

    return sunLight * caustics;
}

void main()
{
    // Setup render target configuration.
#if SSR_QUALITY > 1
    /* DRAWBUFFERS:01 */
#else
    /* DRAWBUFFERS:0 */
#endif

    ivec2 iTexCoords = ivec2(gl_FragCoord.xy);

    float z = texelFetch(depthtex0, iTexCoords, 0).x;
    bool hand = z < 0.6;

    // Early exit for sky/background when depth is at the far plane.
    if (z >= 1.0)
    {
        vec3 color = texture(colortex8, texcoord).rgb;
#ifndef IS_IRIS
        if (!hand)
            color += texture(colortex2, texcoord).rgb;
#endif
        gl_FragData[0].rgb = color;
        return;
    }

    vec3 albedo = texelFetch(colortex0, iTexCoords, 0).rgb;
    vec4 basicTex = texelFetch(colortex2, iTexCoords, 0);
#ifdef PBR
    vec4 extendedTex = texelFetch(colortex3, iTexCoords, 0);
#endif

    vec3 normal = decodeNormal(basicTex.rg);

    // Decode and adjust lightmap coordinates from the blue channel.
    vec2 lightmap = decodeVec2(basicTex.b);
    lightmap.x = lightmap.x * lightmap.x * lightmap.x;

    // Decode additional properties from the alpha channel.
    vec2 decoded = decodeVec2(basicTex.a);
    float ao = decoded.x;
    float sss = max((decoded.y - 0.252941) / 0.747059, 0.0) * 0.75;
    float porosity = min(decoded.y / 0.252941, 1.0);

#ifdef LPV_ENABLED
    // Use an arithmetic average instead of mix for ambient occlusion.
    ao = (ao + 1.0) * 0.5;
#endif

    // Extended material properties for PBR.
    vec3 reflectivity;
    float roughness;
#ifdef PBR
    vec2 decoded2 = decodeVec2(extendedTex.x);

    float rainMult = sqrt(lightmap.y) * wetness * (1.0 - porosity * porosity);
    roughness = mix((1.0 - decoded2.y) * (1.0 - decoded2.y), 0.01, rainMult);
    reflectivity = vec3(decoded2.x);
    
#else
    reflectivity = vec3(0.0);
    roughness = 1.0;
#endif

    // Compute fragment position in screen space.
    vec3 fragpos;
#ifdef TAA
    fragpos = toScreenSpace(vec3(texcoord - (offsets[framemod8] * texelSize * 0.5), z));
#else
    fragpos = toScreenSpace(vec3(texcoord, z));
#endif

    // Set light position (overwritten if not in OVERWORLD mode).
    vec3 lightPosition = lightPos;
#ifndef OVERWORLD
    lightPosition = vec3(0.0, 1.0, 0.0);
#endif

    // Compute world-space position p3.
    vec3 p3 = mat3(gbufferModelViewInverse) * fragpos + gbufferModelViewInverse[3].xyz;

    // Initialize lighting variables.
    vec3 colored = vec3(0.0);
    vec3 sAlbedo = vec3(0.6) * ambientA * lightmap.y;
    float entity = texture(colortex5, texcoord).g;
    float diffuseSun = max(dot(normal, lightPosition), 0.0);
    vec3 specular = vec3(0.0);

    // Lighting calculations.
    if (!hand)
    {
#if SSAO > 0
        ao *= rtao(normal, fragpos);
#else
        ao = pow(ao, 2.0);
#endif

#ifdef OVERWORLD
    #ifdef SHADOWS_ENABLED
        diffuseSun = shadowFunc(diffuseSun, sss, p3, noise_standard(gl_FragCoord.xy), lightmap, fragpos, sAlbedo, albedo, normal, colored, entity);
    #else
        diffuseSun = shadowFunc_no_shadow(diffuseSun, sss, p3, lightmap, ao, fragpos);
    #endif
#endif
    }
    else
    {
        reflectivity = vec3(0.0);
        diffuseSun = max(dot(viewToWorld(normal), lightPosition), 0.0);
        diffuseSun *= min(eyeBrightnessSmooth.y * recip * lightmap.y, 1.0);
    }

    // Torch lighting and ambient light.
    vec3 torchLighting = CalculateTorch(lightmap, normal, fragpos, ao);
    vec3 ambient = calcAmbientLight(normal, lightmap.xy, ambientUp, ambientDown, ambientRight, ambientLeft, ambientB, ambientF, sAlbedo + colored, diffuseSun, torchLighting) * ao;

#ifdef CAUSTICS
    if (!hand)
    {
        vec3 fragposition = mat3(gbufferModelViewInverse) * fragpos + gbufferModelViewInverse[3].xyz +
                            (cameraPosition - previousCameraPosition);
        vec3 previousPosition = mat3(gbufferPreviousModelView) * fragposition +
                                gbufferPreviousModelView[3].xyz;
        previousPosition = toClipSpace3Prev(previousPosition);

        vec2 velocity = previousPosition.xy - texcoord;
        previousPosition.xy = texcoord + velocity;

        float prevmask = texture(colortex4, previousPosition.xy).a;

        bool condition = ((isEyeInWater == 0) && (prevmask > 0.0)) ||
                         ((isEyeInWater == 1) && (prevmask <= 0.0));
        if (condition)
        {
            vec3 caustics = CalculateWaterCaustics(fragpos) * lightmap.y * max(diffuseSun, 0.1);
            ambient += caustics;
        }
    }
#endif

    // Final color composition.
    vec3 color;
#if SSR_QUALITY > 1
    #ifdef HARDCODED_METALS
    color = ambient * albedo;
    #else
    color = ambient * albedo + specular;
    #endif
#else
    color = ambient * albedo + specular;
#endif

#if SSR_QUALITY <= 1
    #ifdef PBR
    if(reflectivity.y > 0)
    color = max(computeReflections(lightmap.y, roughness, reflectivity, albedo,
                                   diffuseSun, normal, p3, ambient * albedo),
                vec3(0.0));
    #endif
#endif
  //   color.rgb = vec3(reflectivity);

    // Write out final fragment color.
    gl_FragData[0].rgb = color;
#if SSR_QUALITY > 1
    gl_FragData[1].rgb = albedo;
    gl_FragData[1].a = diffuseSun;
#endif
}
